#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File  : train_survival.py
# Author: My Duan
# Date  : 2022-09-23

import torch.nn as nn
from survival_utils import *
from PartialNLL import PartialNLL
from torch.autograd import Variable


def l1_regularization(model, l1_alpha):
	l1_loss = []
	for module in model.modules():
		if type(module) is nn.Linear:
			l1_loss.append(torch.abs(module.weight).sum())
	return l1_alpha * sum(l1_loss)


def l2_regularization(model, l2_alpha):
	l2_loss = []
	for module in model.modules():
		if type(module) is nn.Linear:
			l2_loss.append((module.weight ** 2).sum() / 2.0)
	return l2_alpha * sum(l2_loss)


alpha = 0.5

def model_train(model, dataloader, R, optimizer,
				criterion_reg, criterion_cox,
				opt,
				alpha=0.5,
				get_ci=True):

	model.train()

	total_loss = 0.0

	idx = 0
	batch_censored = None
	batch_y = None
	output2 = None

	for idx, batch_loader in enumerate(dataloader):

		batch_t = batch_loader.y[:, :-4].to(opt.device)
		batch_y = batch_loader.y[:, -3].to(opt.device)
		batch_censored = batch_loader.y[:, -4].to(opt.device)
		batch_idx = batch_loader.y[:, -1].cpu().numpy().astype('int')

		try:
			batch_R = R[batch_idx, :][:, batch_idx]
		except:
			batch_R = None
			print('Batch_R Error !!!')

		batch_R = torch.FloatTensor(batch_R).to(opt.device)

		regression, output2 = model(batch_loader)
		loss_reg = criterion_reg(regression, batch_t)
		loss_cox = criterion_cox(output2, batch_R, batch_censored)

		optimizer.zero_grad()

		loss_list = torch.stack([loss_reg, loss_cox]).squeeze()

		if idx == 0:
			initial_task_loss_list = loss_list.detach()

		# alpha = 0.5
		task_list = [1,2]
		loss_ratio_lbtw = loss_list.detach()/initial_task_loss_list
		inverse_traing_rate = loss_ratio_lbtw
		class_weights = inverse_traing_rate.pow(alpha)
		class_weights = class_weights / sum(class_weights) * len(task_list)
		class_weights = Variable(class_weights, requires_grad=False)
		loss_list = torch.mul(loss_list, class_weights)

		loss = loss_list.sum() # return tensor

		loss.backward()
		optimizer.step()

		total_loss += loss_list.sum().detach().item()

	ci = None

	event_observed = np.array([1 if v == 0 else 0 for v in batch_censored])

	if get_ci:
		from lifelines.utils import concordance_index

		batch_y = batch_y.to(opt.device)
		output2 = -output2.reshape(-1).data.to(opt.device)

		ci = concordance_index(batch_y, output2, event_observed)

	return total_loss / (idx + 1), ci


def model_valid(model, dataloader, R, criterion_reg, criterion_cox, alpha, opt, get_ci=True):

	model.eval()

	total_loss = 0.0

	idx = 0
	batch_censored = None
	batch_y = None
	output2 = None


	for idx, batch_loader in enumerate(dataloader):

		batch_t = batch_loader.y[:, :-4].to(opt.device)
		batch_y = batch_loader.y[:, -3].to(opt.device)
		batch_censored = batch_loader.y[:, -4].to(opt.device)
		batch_idx = batch_loader.y[:, -1].cpu().numpy().astype('int')


		try:
			batch_R = R[batch_idx, :][:, batch_idx]
		except:
			batch_R = None
			print('Batch_R Error !!!')

		batch_R = torch.FloatTensor(batch_R).to(opt.device)
		regression, output2 = model(batch_loader)

		loss_reg = criterion_reg(regression, batch_t)
		loss_cox = criterion_cox(output2, batch_R, batch_censored)

		loss_list = torch.stack([loss_reg, loss_cox]).squeeze()

		if idx == 0:
			initial_task_loss_list = loss_list.detach()

		# alpha = 0.5
		task_list = [1, 2]
		loss_ratio_lbtw = loss_list.detach() / initial_task_loss_list
		inverse_traing_rate = loss_ratio_lbtw
		class_weights = inverse_traing_rate.pow(alpha)
		class_weights = class_weights / sum(class_weights) * len(task_list)
		class_weights = Variable(class_weights, requires_grad=False)
		loss_list = torch.mul(loss_list, class_weights)

		total_loss = loss_list.sum().detach().item()  # return tensor

	event_observed = np.array([1 if v == 0 else 0 for v in batch_censored])

	ci = None

	if get_ci:
		from lifelines.utils import concordance_index

		batch_y = batch_y.to(opt.device)
		output2 = -output2.reshape(-1).data.to(opt.device)
		ci = concordance_index(batch_y, output2, event_observed)

	return total_loss / (idx + 1), ci, regression.detach().cpu().numpy()


def train_evaluate(train_loader, valid_loader, test_loader,
					  tr_R, va_R, te_R,
					  model, opt, prefix, alpha, lr, weight_decay, saved_flag=False):

	optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)

	criterion_reg = nn.MSELoss().to(opt.device)
	criterion_cox = PartialNLL().to(opt.device)
	epoch_num = opt.num_epochs

	stopper = EarlyStopper(num_trials=5, model=model,
						   save_model_path=prefix + '/survival_model.pth',
						   save_param_path=prefix + '/survival_param.pth')

	for epoch in range(epoch_num):

		train_loss, tr_ci = model_train(model, train_loader, tr_R,
										optimizer, criterion_reg, criterion_cox, opt,
										alpha=alpha, get_ci=True)

		# print('Epoch is {} :, train_loss is {:.6f}.'.format(epoch, train_loss))

		valid_loss, va_ci, _ = model_valid(model,
										valid_loader, va_R,
										criterion_reg,
										criterion_cox,
										opt=opt,
										alpha=alpha,
										get_ci=True)

		print('Epoch is {} :,\t train_loss:{:.4f},\t valid_loss: {:.4f}'.format(epoch, train_loss, valid_loss))

		if not stopper.is_continuable(model, valid_loss, saved_flag=saved_flag):
			if saved_flag == True:
				print('Model has been saved in {0}!'.format(prefix))
			# print('epoch')
			break


	print('testing ...')
	test_loss, te_ci, te_reg = model_valid(stopper.best_model,
								test_loader, te_R,
								criterion_reg,
								criterion_cox,
								opt=opt,
								alpha=alpha,
								get_ci=True)


	print('test_loss is : {:.6f}, test c_index: {:.6f}'.format(test_loss, te_ci))

	'''get the latent features'''


	for idx, batch_loader in enumerate(test_loader):
		# print(idx)
		x, adj, batch = batch_loader.x.to(opt.device), batch_loader.edge_index.to(opt.device), batch_loader.batch.to(opt.device)

	h_latent = stopper.best_model.health_model(x, adj).detach().cpu().numpy()
	h_latent = h_latent.reshape(te_R.shape[0], -1)
	pd.DataFrame(h_latent).to_csv(prefix + '/HealthM_latent.csv')

	c_latent = stopper.best_model.GAT(x, adj).detach().cpu().numpy()
	c_latent = c_latent.reshape(te_R.shape[0], -1)
	pd.DataFrame(c_latent).to_csv(prefix + '/DiseaseM_latent.csv')

	print('Latent features constructed by HealthModel and DiseaseModel have been saved in {}.'.format(prefix))

	return test_loss, te_ci, te_reg
